@@ -2,16 +2,18 @@
 
 -- predictability
 SET synchronous_commit = on;
+SET extra_float_digits = 0;
 
 DROP TABLE IF EXISTS xpto;
+DROP SEQUENCE IF EXISTS xpto_rand_seq;
 
-SELECT setseed(0);
+CREATE SEQUENCE xpto_rand_seq START 11 INCREMENT 997;
 CREATE TABLE xpto (
 id				serial primary key,
 toasted_col1	text,
-rand1			float8 DEFAULT random(),
+rand1			float8 DEFAULT nextval('xpto_rand_seq'),
 toasted_col2	text,
-rand2 float8	DEFAULT random()
+rand2			float8 DEFAULT nextval('xpto_rand_seq')
 );
 
 SELECT 'init' FROM pg_create_logical_replication_slot('regression_slot', 'wal2json');
@@ -27,11 +29,8 @@ UPDATE xpto SET toasted_col1 = (SELECT string_agg(g.i::text, '') FROM generate_s
 
 UPDATE xpto SET rand1 = 123.456 WHERE id = 1;
 
-SELECT data FROM pg_logical_slot_get_changes('regression_slot', NULL, NULL, 'pretty-print', '1', 'include-typmod', '0', 'include-unchanged-toast', '0');
-
-UPDATE xpto SET rand1 = 234.567 WHERE id = 1;
-
--- include-unchanged-toast=1 is the default
-SELECT data FROM pg_logical_slot_get_changes('regression_slot', NULL, NULL, 'pretty-print', '1', 'include-typmod', '0');
+DELETE FROM xpto WHERE id = 1;
 
+SELECT data FROM pg_logical_slot_peek_changes('regression_slot', NULL, NULL, 'format-version', '1', 'pretty-print', '1', 'include-typmod', '0');
+SELECT data FROM pg_logical_slot_peek_changes('regression_slot', NULL, NULL, 'format-version', '2');
 SELECT 'stop' FROM pg_drop_replication_slot('regression_slot');
