@@ -3,7 +3,7 @@
  * wal2json.c
  * 		JSON output plugin for changeset extraction
  *
- * Copyright (c) 2013-2022, Euler Taveira de Oliveira
+ * Copyright (c) 2013-2024, Euler Taveira de Oliveira
  *
  * IDENTIFICATION
  *		contrib/wal2json/wal2json.c
@@ -20,7 +20,7 @@
 #include "catalog/pg_type.h"
 
 #include "replication/logical.h"
-#if	PG_VERSION_NUM >= 90500
+#if PG_VERSION_NUM >= 90500
 #include "replication/origin.h"
 #endif
 
@@ -34,6 +34,9 @@
 #include "utils/rel.h"
 #include "utils/syscache.h"
 
+#define WAL2JSON_VERSION				"2.6"
+#define WAL2JSON_VERSION_NUM			206
+
 #define	WAL2JSON_FORMAT_VERSION			2
 #define	WAL2JSON_FORMAT_MIN_VERSION		1
 
@@ -69,6 +72,7 @@ typedef struct
 
 	bool		pretty_print;		/* pretty-print JSON? */
 	bool		write_in_chunks;	/* write in chunks? (v1) */
+	bool		numeric_data_types_as_string;	/* use strings for numeric data types */
 
 	JsonAction	actions;			/* output only these actions */
 
@@ -120,16 +124,16 @@ static void pg_decode_commit_txn(LogicalDecodingContext *ctx,
 static void pg_decode_change(LogicalDecodingContext *ctx,
 				 ReorderBufferTXN *txn, Relation rel,
 				 ReorderBufferChange *change);
-#if	PG_VERSION_NUM >= 90500
+#if PG_VERSION_NUM >= 90500
 static bool pg_filter_by_origin(LogicalDecodingContext *ctx, RepOriginId origin_id);
 #endif
-#if	PG_VERSION_NUM >= 90600
+#if PG_VERSION_NUM >= 90600
 static void pg_decode_message(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, XLogRecPtr lsn,
 					bool transactional, const char *prefix,
 					Size content_size, const char *content);
 #endif
-#if	PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000
 static void pg_decode_truncate(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, int n, Relation relations[],
 					ReorderBufferChange *change);
@@ -156,13 +160,13 @@ static void pg_decode_commit_txn_v1(LogicalDecodingContext *ctx,
 static void pg_decode_change_v1(LogicalDecodingContext *ctx,
 				 ReorderBufferTXN *txn, Relation rel,
 				 ReorderBufferChange *change);
-#if	PG_VERSION_NUM >= 90600
+#if PG_VERSION_NUM >= 90600
 static void pg_decode_message_v1(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, XLogRecPtr lsn,
 					bool transactional, const char *prefix,
 					Size content_size, const char *content);
 #endif
-#if	PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000
 static void pg_decode_truncate_v1(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, int n, Relation relations[],
 					ReorderBufferChange *change);
@@ -179,13 +183,13 @@ static void pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN
 static void pg_decode_change_v2(LogicalDecodingContext *ctx,
 				 ReorderBufferTXN *txn, Relation rel,
 				 ReorderBufferChange *change);
-#if	PG_VERSION_NUM >= 90600
+#if PG_VERSION_NUM >= 90600
 static void pg_decode_message_v2(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, XLogRecPtr lsn,
 					bool transactional, const char *prefix,
 					Size content_size, const char *content);
 #endif
-#if	PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000
 static void pg_decode_truncate_v2(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, int n, Relation relations[],
 					ReorderBufferChange *change);
@@ -204,6 +208,12 @@ static void pg_decode_truncate_v2(LogicalDecodingContext *ctx,
 #endif
 #endif
 
+#if PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+static void update_replication_progress(LogicalDecodingContext *ctx, bool skipped_xact);
+#elif PG_VERSION_NUM >= 100000 && PG_VERSION_NUM < 150000
+static void update_replication_progress(LogicalDecodingContext *ctx);
+#endif
+
 void
 _PG_init(void)
 {
@@ -220,13 +230,13 @@ _PG_output_plugin_init(OutputPluginCallbacks *cb)
 	cb->change_cb = pg_decode_change;
 	cb->commit_cb = pg_decode_commit_txn;
 	cb->shutdown_cb = pg_decode_shutdown;
-#if	PG_VERSION_NUM >= 90500
+#if PG_VERSION_NUM >= 90500
 	cb->filter_by_origin_cb = pg_filter_by_origin;
 #endif
-#if	PG_VERSION_NUM >= 90600
+#if PG_VERSION_NUM >= 90600
 	cb->message_cb = pg_decode_message;
 #endif
-#if	PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000
 	cb->truncate_cb = pg_decode_truncate;
 #endif
 }
@@ -261,6 +271,7 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 	data->include_typmod = true;
 	data->include_domain_data_type = false;
 	data->include_column_positions = false;
+	data->numeric_data_types_as_string = false;
 	data->pretty_print = false;
 	data->write_in_chunks = false;
 	data->include_lsn = false;
@@ -480,6 +491,19 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
 							 strVal(elem->arg), elem->defname)));
 		}
+		else if (strcmp(elem->defname, "numeric-data-types-as-string") == 0)
+		{
+			if (elem->arg == NULL)
+			{
+				elog(DEBUG1, "numeric-data-types-as-string argument is null");
+				data->numeric_data_types_as_string = true;
+			}
+			else if (!parse_bool(strVal(elem->arg), &data->numeric_data_types_as_string))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("could not parse value \"%s\" for parameter \"%s\"",
+							 strVal(elem->arg), elem->defname)));
+		}
 		else if (strcmp(elem->defname, "pretty-print") == 0)
 		{
 			if (elem->arg == NULL)
@@ -887,9 +911,39 @@ pg_decode_commit_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
-#if PG_VERSION_NUM >= 150000
-    OutputPluginUpdateProgress(ctx, false);
-#elif PG_VERSION_NUM >= 100000
+	/*
+	 * Some older minor versions from back branches (10 to 14) calls
+	 * OutputPluginUpdateProgress(). That's before the fix
+	 * f95d53eded55ecbf037f6416ced6af29a2c3caca. After that,
+	 * update_replication_progress() function is used for back branches. In
+	 * version 15, update_replication_progress() changes the signature to
+	 * support skipped transactions. In version 16,
+	 * OutputPluginUpdateProgress() is back because a proper fix was added into
+	 * logical decoding.
+	 */
+#if PG_VERSION_NUM >= 160000
+	OutputPluginUpdateProgress(ctx, false);		/* XXX change 2nd param when skipped empty transaction is supported */
+#elif PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+	update_replication_progress(ctx, false);	/* XXX change 2nd param when skipped empty transaction is supported */
+#elif PG_VERSION_NUM >= 140004 && PG_VERSION_NUM < 150000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 130008 && PG_VERSION_NUM < 140000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 120012 && PG_VERSION_NUM < 130000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 110017 && PG_VERSION_NUM < 120000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 100022 && PG_VERSION_NUM < 110000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 140000 && PG_VERSION_NUM < 140004
+	OutputPluginUpdateProgress(ctx);
+#elif PG_VERSION_NUM >= 130000 && PG_VERSION_NUM < 130008
+	OutputPluginUpdateProgress(ctx);
+#elif PG_VERSION_NUM >= 120000 && PG_VERSION_NUM < 120012
+	OutputPluginUpdateProgress(ctx);
+#elif PG_VERSION_NUM >= 110000 && PG_VERSION_NUM < 110017
+	OutputPluginUpdateProgress(ctx);
+#elif PG_VERSION_NUM >= 100000 && PG_VERSION_NUM < 100022
 	OutputPluginUpdateProgress(ctx);
 #endif
 
@@ -1189,7 +1243,7 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 				SysScanDesc			scan;
 				HeapTuple			def_tuple;
 				Datum				def_value;
-				bool				isnull;
+				bool				attisnull;
 				char				*result;
 
 				ScanKeyInit(&scankeys[0],
@@ -1207,9 +1261,9 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 				def_tuple = systable_getnext(scan);
 				if (HeapTupleIsValid(def_tuple))
 				{
-					def_value = fastgetattr(def_tuple, Anum_pg_attrdef_adbin, defrel->rd_att, &isnull);
+					def_value = fastgetattr(def_tuple, Anum_pg_attrdef_adbin, defrel->rd_att, &attisnull);
 
-					if (!isnull)
+					if (!attisnull)
 					{
 						result = TextDatumGetCString(DirectFunctionCall2(pg_get_expr,
 																	def_value,
@@ -1259,6 +1313,10 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 			 *
 			 * The NaN and Infinity are not valid JSON symbols. Hence,
 			 * regardless of sign they are represented as the string null.
+			 *
+			 * Exception to this is when data->numeric_data_types_as_string is
+			 * true. In this case, numbers (including NaN and Infinity values)
+			 * are printed with quotes.
 			 */
 			switch (typid)
 			{
@@ -1269,7 +1327,18 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 				case FLOAT4OID:
 				case FLOAT8OID:
 				case NUMERICOID:
-					if (pg_strncasecmp(outputstr, "NaN", 3) == 0 ||
+					if (data->numeric_data_types_as_string) {
+						if (strspn(outputstr, "0123456789+-eE.") == strlen(outputstr) ||
+								pg_strncasecmp(outputstr, "NaN", 3) == 0 ||
+								pg_strncasecmp(outputstr, "Infinity", 8) == 0 ||
+								pg_strncasecmp(outputstr, "-Infinity", 9) == 0) {
+							appendStringInfo(&colvalues, "%s", comma);
+							escape_json(&colvalues, outputstr);
+						} else {
+							elog(ERROR, "%s is not a number", outputstr);
+						}
+					}
+					else if (pg_strncasecmp(outputstr, "NaN", 3) == 0 ||
 							pg_strncasecmp(outputstr, "Infinity", 8) == 0 ||
 							pg_strncasecmp(outputstr, "-Infinity", 9) == 0)
 					{
@@ -1598,6 +1667,20 @@ pg_decode_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
+#if PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+	update_replication_progress(ctx, false);
+#elif PG_VERSION_NUM >= 140004 && PG_VERSION_NUM < 150000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 130008 && PG_VERSION_NUM < 140000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 120012 && PG_VERSION_NUM < 130000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 110017 && PG_VERSION_NUM < 120000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 100022 && PG_VERSION_NUM < 110000
+	update_replication_progress(ctx);
+#endif
+
 	if (data->format_version == 2)
 		pg_decode_change_v2(ctx, txn, relation, change);
 	else if (data->format_version == 1)
@@ -1764,7 +1847,7 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	appendStringInfo(ctx->out, ",%s", data->nl);
 
 	if (data->include_pk)
-#if	PG_VERSION_NUM >= 100000
+#if PG_VERSION_NUM >= 100000
 		pkbs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_PRIMARY_KEY);
 #else
 		pkbs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_KEY);
@@ -1774,23 +1857,51 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	{
 		case REORDER_BUFFER_CHANGE_INSERT:
 			/* Print the new tuple */
+#if	PG_VERSION_NUM >= 100000
+			if (data->include_pk && OidIsValid(relation->rd_pkindex))
+#else
 			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
 					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+#endif
 			{
+#if	PG_VERSION_NUM >= 170000
+				columns_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, true, relation);
+				pk_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, pkbs, false);
+#else
 				columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, relation);
 				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, false);
+#endif
 			}
 			else
 			{
+#if	PG_VERSION_NUM >= 170000
+				columns_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, false, relation);
+#else
 				columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, false, relation);
+#endif
 			}
 			break;
 		case REORDER_BUFFER_CHANGE_UPDATE:
 			/* Print the new tuple */
+#if	PG_VERSION_NUM >= 170000
+			columns_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, true, relation);
+#else
 			columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, relation);
+#endif
+
+#if	PG_VERSION_NUM >= 100000
+			if (data->include_pk && OidIsValid(relation->rd_pkindex))
+#else
 			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
 					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+#endif
+			{
+#if	PG_VERSION_NUM >= 170000
+				pk_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, pkbs, true);
+#else
 				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, true);
+#endif
+			}
 
 			/*
 			 * The old tuple is available when:
@@ -1806,21 +1917,43 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 				elog(DEBUG1, "old tuple is null");
 
 				ribs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_IDENTITY_KEY);
+#if	PG_VERSION_NUM >= 170000
+				identity_to_stringinfo(ctx, tupdesc, change->data.tp.newtuple, ribs);
+#else
 				identity_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, ribs);
+#endif
 			}
 			else
 			{
 				elog(DEBUG1, "old tuple is not null");
+#if	PG_VERSION_NUM >= 170000
+				identity_to_stringinfo(ctx, tupdesc, change->data.tp.oldtuple, NULL);
+#else
 				identity_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, NULL);
+#endif
 			}
 			break;
 		case REORDER_BUFFER_CHANGE_DELETE:
+#if	PG_VERSION_NUM >= 100000
+			if (data->include_pk && OidIsValid(relation->rd_pkindex))
+#else
 			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
 					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
-				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, true);
+#endif
+			{
+#if	PG_VERSION_NUM >= 170000
+				pk_to_stringinfo(ctx, tupdesc, change->data.tp.oldtuple, pkbs, true);
+#else
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, pkbs, true);
+#endif
+			}
 
 			ribs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_IDENTITY_KEY);
+#if	PG_VERSION_NUM >= 170000
+			identity_to_stringinfo(ctx, tupdesc, change->data.tp.oldtuple, ribs);
+#else
 			identity_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, ribs);
+#endif
 
 			if (change->data.tp.oldtuple == NULL)
 				elog(DEBUG1, "old tuple is null");
@@ -1846,9 +1979,12 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 static void
 pg_decode_write_value(LogicalDecodingContext *ctx, Datum value, bool isnull, Oid typid)
 {
-	Oid		typoutfunc;
-	bool	isvarlena;
-	char	*outstr;
+	JsonDecodingData	*data;
+	Oid					typoutfunc;
+	bool				isvarlena;
+	char				*outstr;
+
+	data = ctx->output_plugin_private;
 
 	if (isnull)
 	{
@@ -1885,6 +2021,10 @@ pg_decode_write_value(LogicalDecodingContext *ctx, Datum value, bool isnull, Oid
 	 *
 	 * The NaN an Infinity are not valid JSON symbols. Hence, regardless of
 	 * sign they are represented as the string null.
+	 *
+	 * Exception to this is when data->numeric_data_types_as_string is
+	 * true. In this case, numbers (including NaN and Infinity values)
+	 * are printed with quotes.
 	 */
 	switch (typid)
 	{
@@ -1895,7 +2035,17 @@ pg_decode_write_value(LogicalDecodingContext *ctx, Datum value, bool isnull, Oid
 		case FLOAT4OID:
 		case FLOAT8OID:
 		case NUMERICOID:
-			if (pg_strncasecmp(outstr, "NaN", 3) == 0 ||
+			if (data->numeric_data_types_as_string) {
+				if (strspn(outstr, "0123456789+-eE.") == strlen(outstr) ||
+						pg_strncasecmp(outstr, "NaN", 3) == 0 ||
+						pg_strncasecmp(outstr, "Infinity", 8) == 0 ||
+						pg_strncasecmp(outstr, "-Infinity", 9) == 0) {
+					escape_json(ctx->out, outstr);
+				} else {
+					elog(ERROR, "%s is not a number", outstr);
+				}
+			}
+			else if (pg_strncasecmp(outstr, "NaN", 3) == 0 ||
 					pg_strncasecmp(outstr, "Infinity", 8) == 0 ||
 					pg_strncasecmp(outstr, "-Infinity", 9) == 0)
 			{
@@ -1953,7 +2103,7 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 	}
 	else if (kind == PGOUTPUTJSON_PK)
 	{
-#if	PG_VERSION_NUM >= 100000
+#if PG_VERSION_NUM >= 100000
 		bs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_PRIMARY_KEY);
 #else
 		bs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_KEY);
@@ -2259,7 +2409,11 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 	if (change->data.tp.newtuple != NULL)
 	{
 		appendStringInfoString(ctx->out, ",\"columns\":[");
+#if PG_VERSION_NUM >= 170000
+		pg_decode_write_tuple(ctx, relation, change->data.tp.newtuple, PGOUTPUTJSON_CHANGE);
+#else
 		pg_decode_write_tuple(ctx, relation, &change->data.tp.newtuple->tuple, PGOUTPUTJSON_CHANGE);
+#endif
 		appendStringInfoChar(ctx->out, ']');
 	}
 
@@ -2279,7 +2433,11 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 	if (change->data.tp.oldtuple != NULL)
 	{
 		appendStringInfoString(ctx->out, ",\"identity\":[");
+#if	PG_VERSION_NUM >= 170000
+		pg_decode_write_tuple(ctx, relation, change->data.tp.oldtuple, PGOUTPUTJSON_IDENTITY);
+#else
 		pg_decode_write_tuple(ctx, relation, &change->data.tp.oldtuple->tuple, PGOUTPUTJSON_IDENTITY);
+#endif
 		appendStringInfoChar(ctx->out, ']');
 	}
 	else
@@ -2296,7 +2454,7 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 			 * Before v10, there is not rd_pkindex then rely on REPLICA
 			 * IDENTITY DEFAULT to obtain primary key.
 			 */
-#if	PG_VERSION_NUM >= 100000
+#if PG_VERSION_NUM >= 100000
 			if (OidIsValid(relation->rd_pkindex) || OidIsValid(relation->rd_replidindex))
 #else
 			if (OidIsValid(relation->rd_replidindex))
@@ -2304,7 +2462,11 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 			{
 				elog(DEBUG1, "REPLICA IDENTITY: obtain old tuple using new tuple");
 				appendStringInfoString(ctx->out, ",\"identity\":[");
+#if PG_VERSION_NUM >= 170000
+				pg_decode_write_tuple(ctx, relation, change->data.tp.newtuple, PGOUTPUTJSON_IDENTITY);
+#else
 				pg_decode_write_tuple(ctx, relation, &change->data.tp.newtuple->tuple, PGOUTPUTJSON_IDENTITY);
+#endif
 				appendStringInfoChar(ctx->out, ']');
 			}
 			else
@@ -2324,16 +2486,23 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 	if (data->include_pk)
 	{
 		appendStringInfoString(ctx->out, ",\"pk\":[");
-#if	PG_VERSION_NUM >= 100000
+#if PG_VERSION_NUM >= 100000
 		if (OidIsValid(relation->rd_pkindex))
 #else
 		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
 #endif
 		{
+#if PG_VERSION_NUM >= 170000
+			if (change->data.tp.oldtuple != NULL)
+				pg_decode_write_tuple(ctx, relation, change->data.tp.oldtuple, PGOUTPUTJSON_PK);
+			else
+				pg_decode_write_tuple(ctx, relation, change->data.tp.newtuple, PGOUTPUTJSON_PK);
+#else
 			if (change->data.tp.oldtuple != NULL)
 				pg_decode_write_tuple(ctx, relation, &change->data.tp.oldtuple->tuple, PGOUTPUTJSON_PK);
 			else
 				pg_decode_write_tuple(ctx, relation, &change->data.tp.newtuple->tuple, PGOUTPUTJSON_PK);
+#endif
 		}
 		appendStringInfoChar(ctx->out, ']');
 	}
@@ -2386,7 +2555,7 @@ pg_decode_change_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	MemoryContextReset(data->context);
 }
 
-#if	PG_VERSION_NUM >= 90600
+#if PG_VERSION_NUM >= 90600
 /* Callback for generic logical decoding messages */
 static void
 pg_decode_message(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
@@ -2395,6 +2564,20 @@ pg_decode_message(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
+#if PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+	update_replication_progress(ctx, false);
+#elif PG_VERSION_NUM >= 140004 && PG_VERSION_NUM < 150000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 130008 && PG_VERSION_NUM < 140000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 120012 && PG_VERSION_NUM < 130000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 110017 && PG_VERSION_NUM < 120000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 100022 && PG_VERSION_NUM < 110000
+	update_replication_progress(ctx);
+#endif
+
 	/* Filter message prefixes, if available */
 	if (list_length(data->filter_msg_prefixes) > 0)
 	{
@@ -2479,7 +2662,12 @@ pg_decode_message_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 
 	appendStringInfo(ctx->out, "%s%s", data->ht, data->ht);
 
-	if (data->nr_changes > 1)
+	/*
+	 * Non-transactional message contains only one object. Comma is not
+	 * required. Avoid printing a comma for non-transactional messages that was
+	 * provided in a transaction.
+	 */
+	if (transactional && data->nr_changes > 1)
 		appendStringInfoChar(ctx->out, ',');
 
 	appendStringInfo(ctx->out, "{%s%s%s%s\"kind\":%s\"message\",%s", data->nl, data->ht, data->ht, data->ht, data->sp, data->nl);
@@ -2591,7 +2779,7 @@ pg_decode_message_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 }
 #endif
 
-#if	PG_VERSION_NUM >= 110000
+#if PG_VERSION_NUM >= 110000
 /* Callback for TRUNCATE command */
 static void pg_decode_truncate(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn, int n, Relation relations[],
@@ -2599,6 +2787,22 @@ static void pg_decode_truncate(LogicalDecodingContext *ctx,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
+	/*
+	 * For back branches (10 to 15), update_replication_progress() should be called here.
+	 * FIXME call OutputPluginUpdateProgress() for old minor versions?
+	 */
+#if PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+	update_replication_progress(ctx, false);
+#elif PG_VERSION_NUM >= 140004 && PG_VERSION_NUM < 150000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 130008 && PG_VERSION_NUM < 140000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 120012 && PG_VERSION_NUM < 130000
+	update_replication_progress(ctx);
+#elif PG_VERSION_NUM >= 110017 && PG_VERSION_NUM < 120000
+	update_replication_progress(ctx);
+#endif
+
 	if (data->format_version == 2)
 		pg_decode_truncate_v2(ctx, txn, n, relations, change);
 	else if (data->format_version == 1)
@@ -3062,3 +3266,65 @@ split_string_to_oid_list(char *rawstring, char separator, List **sl)
 
 	return true;
 }
+
+/*
+ * Try to update progress and send a keepalive message if too many changes were
+ * processed.
+ *
+ * For a large transaction, if we don't send any change to the downstream for a
+ * long time (exceeds the wal_receiver_timeout of standby) then it can timeout.
+ * This can happen when all or most of the changes are not published.
+ *
+ * Copied from Postgres commit f95d53eded55ecbf037f6416ced6af29a2c3caca
+ */
+#if PG_VERSION_NUM >= 150000 && PG_VERSION_NUM < 160000
+static void
+update_replication_progress(LogicalDecodingContext *ctx, bool skipped_xact)
+{
+	static int	changes_count = 0;
+
+	/*
+	 * We don't want to try sending a keepalive message after processing each
+	 * change as that can have overhead. Tests revealed that there is no
+	 * noticeable overhead in doing it after continuously processing 100 or so
+	 * changes.
+	 */
+#define CHANGES_THRESHOLD 100
+
+	/*
+	 * If we are at the end of transaction LSN, update progress tracking.
+	 * Otherwise, after continuously processing CHANGES_THRESHOLD changes, we
+	 * try to send a keepalive message if required.
+	 */
+	if (ctx->end_xact || ++changes_count >= CHANGES_THRESHOLD)
+	{
+		OutputPluginUpdateProgress(ctx, skipped_xact);
+		changes_count = 0;
+	}
+}
+#elif PG_VERSION_NUM >= 100000 && PG_VERSION_NUM < 150000
+static void
+update_replication_progress(LogicalDecodingContext *ctx)
+{
+	static int	changes_count = 0;
+
+	/*
+	 * We don't want to try sending a keepalive message after processing each
+	 * change as that can have overhead. Tests revealed that there is no
+	 * noticeable overhead in doing it after continuously processing 100 or so
+	 * changes.
+	 */
+#define CHANGES_THRESHOLD 100
+
+	/*
+	 * If we are at the end of transaction LSN, update progress tracking.
+	 * Otherwise, after continuously processing CHANGES_THRESHOLD changes, we
+	 * try to send a keepalive message if required.
+	 */
+	if (ctx->end_xact || ++changes_count >= CHANGES_THRESHOLD)
+	{
+		OutputPluginUpdateProgress(ctx);
+		changes_count = 0;
+	}
+}
+#endif
